using System;
using UnityEngine;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using SimpleJSON;

namespace Vinput
{


    public class CtrlSnapshot
    {
        public string name = string.Empty;
        public string currentPositionState = string.Empty;
        public string currentRotationState = string.Empty;
        public float RBHoldPositionSpring = 2000;
        public float RBHoldRotationSpring = 250;
        public Vector3 localPosition = Vector3.zero;
        public Quaternion localRotation = Quaternion.identity;
        public Matrix4x4 worldMatrix = Matrix4x4.identity;

        public CtrlSnapshot(FreeControllerV3 control = null)
        {
            if (control == null) return;

            name = control.name;
            currentPositionState = control.currentPositionState.ToString();
            currentRotationState = control.currentRotationState.ToString();
            RBHoldPositionSpring = control.RBHoldPositionSpring;
            RBHoldRotationSpring = control.RBHoldRotationSpring;
            localPosition = control.transform.localPosition;
            localRotation = control.transform.localRotation;
            worldMatrix = control.transform.localToWorldMatrix;
        }
    }
    public class Cycler
    {
        public bool __DEBUG__ = false;

        private bool _isActive = false;

        private float _currTime = 0.0f;
        private Coroutine _co_instance = null;

        public bool IsActive => _isActive;
        public float CurrentTime => _currTime;
        public float CurrentPct => _currTime / CurrentDuration;
        public float CurrentT => T(_currTime / CurrentDuration);
        public float CurrentDuration = 1.0f;
        private readonly Action<Cycler> _onBegin;
        private readonly Action<Cycler> _onUpdate;
        private readonly Action<Cycler> _onEnd;

        private readonly MonoBehaviour _mono;

        private JSONStorableFloat _cycleDurationJSON;
        private JSONStorableFloat _cyclePctInJSON;
        private JSONStorableFloat _cycleDurationRandomJSON;
        private JSONStorableStringChooser _easingInJSON;
        private JSONStorableStringChooser _easingOutJSON;

        public Cycler(MonoBehaviour mono, Action<Cycler> OnBegin = null, Action<Cycler> OnUpdate = null, Action<Cycler> OnEnd = null)
        {
            _mono = mono;
            _onBegin = OnBegin ?? ((_) => { });
            _onUpdate = OnUpdate ?? ((_) => { });
            _onEnd = OnEnd ?? ((_) => { });
        }

        public void StartCycle()
        {
            if (_co_instance == null)
            {
                _currTime = 0.0f;
                _isActive = true;
                _co_instance = _mono.StartCoroutine(_cycle());
                if (__DEBUG__) SuperController.LogMessage($"Starting cycle... (_co_instance = {_co_instance})");
            }
        }

        public void StopCycle()
        {
            if (__DEBUG__) SuperController.LogMessage($"Stopping cycle... (_co_instance = {_co_instance})");
            if (_co_instance == null) return;

            _mono.StopCoroutine(_co_instance);
            _co_instance = null;
            _isActive = false;
        }

        public float T(float currPct)
        {
            bool r = currPct >= _cyclePctInJSON.val;
            float x = r ? (currPct - _cyclePctInJSON.val) / (1 - _cyclePctInJSON.val) : currPct / _cyclePctInJSON.val;
            return r ? 1 - Easings.EasingFuncs[_easingOutJSON.val](x) : Easings.EasingFuncs[_easingInJSON.val](x);
        }

        private IEnumerator _cycle()
        {
            CurrentDuration = Math.Max(UnityEngine.Random.Range(_cycleDurationJSON.val * (1 - _cycleDurationRandomJSON.val), _cycleDurationJSON.val * (1 + _cycleDurationRandomJSON.val)), 0.1f);
            _onBegin(this);

            while (_currTime <= CurrentDuration)
            {
                if (__DEBUG__) SuperController.LogMessage($"Calling OnUpdate (time = {_currTime}, pct = {_currTime / CurrentDuration})");
                _onUpdate(this);
                _currTime += Time.deltaTime;
                yield return null;
            }

            _endCycle();
        }

        private void _endCycle()
        {
            _co_instance = null;
            _onEnd(this);

            if (_isActive) StartCycle();
        }

        public void SetupUI(MVRScript script)
        {
            _cycleDurationJSON = UI_Utils.SetupSlider(script, "Cycle Duration", 1.0f, 0.1f, 5.0f, false);
            _cycleDurationRandomJSON = UI_Utils.SetupSlider(script, "Duration Random %", 0.0f, 0.0f, 0.5f);
            _cyclePctInJSON = UI_Utils.SetupSlider(script, "Duration % In", 0.5f, 0.1f, 0.9f);
            
            _easingInJSON = new JSONStorableStringChooser("EaseIn", Easings.EasingTypes, "Linear", "Ease In");
            script.CreateScrollablePopup(_easingInJSON);
            script.RegisterStringChooser(_easingInJSON);

            _easingOutJSON = new JSONStorableStringChooser("EaseOut", Easings.EasingTypes, "Linear", "Ease Out");
            script.CreateScrollablePopup(_easingOutJSON);
            script.RegisterStringChooser(_easingOutJSON);
        }

    }

    public static class Easings
    {
        // Constants for easing functions
        private static readonly float BACK_C1 = 1.70158f;
        private static readonly float BACK_C2 = BACK_C1 + 1;
        private static readonly float BACK_C3 = BACK_C1 * 1.525f;
        private static readonly float BOUNCE_C1 = 7.5625f;
        private static readonly float BOUNCE_C2 = 2.75f;

        // Easing functions
        public static Dictionary<string, Func<float, float>> EasingFuncs = new Dictionary<string, Func<float, float>>
    {
        // Reference: https://easings.net/
        {"Linear", f => f},
        {"Quadratic In", f => (float)Math.Pow(f, 2) },
        {"Quadratic Out", f => (float)(1 - Math.Pow(1 - f, 2)) },
        {"Quadratic InOut", f => f < 0.5 ? (float)(2 * Math.Pow(f, 2)) : (float)(1 - Math.Pow(2 * (1 - f), 2) / 2) },
        {"Cubic In", f => (float)Math.Pow(f, 3) },
        {"Cubic Out", f => (float)(1 - Math.Pow(1 - f, 3)) },
        {"Cubic InOut", f => f < 0.5 ? (float)(4 * Math.Pow(f, 3)) : (float)(1 - Math.Pow(2 * (1 - f), 3) / 2) },
        {"Sinusoidal In", f => (float)(1 - Math.Cos(f * Math.PI / 2)) },
        {"Sinusoidal Out", f => (float)Math.Sin(f * Math.PI / 2) },
        {"Sinusoidal InOut", f => -(float)(Math.Cos(Math.PI * f) - 1) / 2 },
        {"Exponential In", f => f == 0 ? 0 : (float)Math.Pow(2, 10 * f - 10) },
        {"Exponential Out", f => f == 1 ? 1 : (float)(1 - Math.Pow(2, -10 * f)) },
        {"Exponential InOut", f => f == 0 || f == 1 ? f :  f < 0.5 ? (float)Math.Pow(2, 20 * f - 10) / 2 : (float)(2 - Math.Pow(2, -20 * f + 10)) / 2 },
        {"Back In", f => (float)(BACK_C2 * Math.Pow(f, 3) - BACK_C1 * Math.Pow(f, 2)) },
        {"Back Out", f => (float)(1 + BACK_C2 * Math.Pow(f - 1, 3) + BACK_C1 * Math.Pow(f - 1, 2)) },
        {"Back InOut", f => f < 0.5 ? (float)(Math.Pow(2 * f, 2) * ((BACK_C3 + 1) * 2 * f - BACK_C3)) / 2 : (float)(Math.Pow(2 * f - 2, 2) * ((BACK_C3 + 1) * (f * 2 - 2) + BACK_C3) + 2) / 2 },
        {"Bounce In", f => 1 - _bounceOut(1 - f) },
        {"Bounce Out", _bounceOut },
        {"Bounce InOut", f => f < 0.5 ? (1 - _bounceOut(1 - 2*f)) / 2 : (1 + _bounceOut(2*f - 1)) / 2 }
    };

        private static float _bounceOut(float f)
        {
            if (f < 1 / BOUNCE_C2)
                return BOUNCE_C1 * f * f;

            if (f < 2 / BOUNCE_C2)
                return BOUNCE_C1 * (f -= 1.5f / BOUNCE_C2) * f + 0.75f;

            if (f < 2.5 / BOUNCE_C2)
                return BOUNCE_C1 * (f -= 2.25f / BOUNCE_C2) * f + 0.9375f;

            return BOUNCE_C1 * (f -= 2.625f / BOUNCE_C2) * f + 0.984375f;
        }

        public static List<string> EasingTypes => EasingFuncs.Keys.ToList();
    }


    public class GensAligner
    {
        private bool _isEnabled;

        private readonly TargetManager _tgtMgr;

        private readonly Atom _hostAtom;

        private readonly List<FreeControllerV3> _ctrls = new List<FreeControllerV3> { };
        private readonly Rigidbody _pelvisRB;

        private readonly List<CtrlSnapshot> _ctrlSnapshots = new List<CtrlSnapshot> { };
        private static readonly List<string> _ctrlNames = new List<string> { "penisBaseControl", "penisMidControl", "penisTipControl" };

        public bool IsEnabled => _isEnabled;
        private bool _isValid => _ctrls.Count == 3 && _tgtMgr.TargetRB1 != null && _tgtMgr.TargetRB2 != null;

        public GensAligner(Atom hostAtom, TargetManager targetManager)
        {
            _hostAtom = hostAtom;
            _tgtMgr = targetManager;

            if (_hostAtom != null)
            {
                _ctrlNames.ForEach(cname =>
                {
                    FreeControllerV3 ctl = _hostAtom.freeControllers.First(fc => fc.name == cname);
                    _ctrls.Add(ctl);
                });

                _pelvisRB = _hostAtom.rigidbodies.First(rb => rb.name == "pelvis");

            }

        }

        public void Enable()
        {
            if (_isValid)
            {
                if (!_ctrlSnapshots.Any())
                {
                    _ctrls.ForEach(ctrl =>
                    {
                        _ctrlSnapshots.Add(new CtrlSnapshot(ctrl));
                    });
                }

                _ctrls.ForEach(ctrl =>
                    {
                        ctrl.RBHoldRotationSpring = 1000;
                        ctrl.currentRotationState = FreeControllerV3.RotationState.On;
                        if (ctrl.name != "penisBaseControl") ctrl.currentPositionState = FreeControllerV3.PositionState.Off;
                    });

                _isEnabled = true;
            }
        }

        public void Disable()
        {
            if (_isEnabled)
            {
                _isEnabled = false;
                for (int i = 0; i < _ctrlSnapshots.Count; i++)
                {
                    _ctrls[i].currentRotationState = (FreeControllerV3.RotationState)Enum.Parse(typeof(FreeControllerV3.RotationState), _ctrlSnapshots[i].currentRotationState);
                    _ctrls[i].transform.localRotation = _ctrlSnapshots[i].localRotation;
                    _ctrls[i].RBHoldRotationSpring = _ctrlSnapshots[i].RBHoldRotationSpring;
                }

                _ctrlSnapshots.Clear();
            };
        }

        public void Update()
        {
            if (_isEnabled)
            {
                Vector3 upVec = _pelvisRB.transform.up;
                Vector3 pos1 = _tgtMgr.TargetPos1;
                Vector3 pos2 = _tgtMgr.TargetPos2;

                _ctrls.ForEach(ctrl =>
                {
                    Vector3 ctrlpos = ctrl.transform.position;

                    if (Vector3.Distance(ctrlpos, pos2) < 0.05) return;

                    if (Vector3.Distance(ctrlpos, pos1) < 0.02 ||
                     Vector3.Distance(ctrlpos, pos2) <= Vector3.Distance(pos1, pos2))
                        ctrl.transform.LookAt(pos2, upVec);

                    else ctrl.transform.LookAt(Vector3.Lerp(pos2, pos1, (Vector3.Distance(pos1, ctrlpos) - 0.02f) / 0.1f), upVec);
                });
            }
        }

        public JSONNode GetJSON()
        {
            if (_ctrlSnapshots.Any())
            {
                JSONArray arr = new JSONArray();

                _ctrlSnapshots.ForEach(snapshot =>
                {
                    arr.Add(Serializer.Serialize(snapshot));
                });

                return arr;
            }

            return null;
        }

        public void RestoreFromJSON(JSONNode jc)
        {

            if (jc.AsArray.Count > 0)
            {
                _ctrlSnapshots.Clear();
                foreach (JSONClass snapshot in jc.AsArray)
                {
                    _ctrlSnapshots.Add(Serializer.DeserializeCtrlSnapshot(snapshot));
                }
            }
        }

    }

    public static class Scene_Utils
    {
        public static Vector3 PositionFromMatrix(Matrix4x4 mat)
        {
            if (mat == null) return Vector3.zero;

            return new Vector3(mat[0, 3], mat[1, 3], mat[2, 3]);
        }
        public static List<string> ControlNames(Atom hostAtom)
        {
            if (hostAtom == null) return new List<string> { };

            return hostAtom.freeControllers.Select(c => c.name).ToList(); ;
        }

        public static List<string> PersonNames => SuperController.singleton.GetAtoms().Where(atom => atom.category == "People").Select(atom => atom.name).ToList();
        public static List<string> GetAtomNames()
        {
            List<string> lst = SuperController.singleton.GetAtoms().Select(atom => atom.name).ToList();
            lst.Sort();
            return lst;
        }
    }

    public static class Serializer
    {
        public static JSONClass Serialize(Matrix4x4 mat)
        {
            if (mat == null) mat = Matrix4x4.identity;

            var jc = new JSONClass
            {
                ["0"] = Serialize(mat.GetColumn(0)),
                ["1"] = Serialize(mat.GetColumn(1)),
                ["2"] = Serialize(mat.GetColumn(2)),
                ["3"] = Serialize(mat.GetColumn(3))
            };
            return jc;
        }
        public static JSONClass Serialize(Quaternion rot)
        {
            return Serialize(new Vector4(rot.x, rot.y, rot.z, rot.w));
        }

        public static JSONClass Serialize(Vector3 pos)
        {
            var jc = new JSONClass
            {
                ["x"] = { AsFloat = pos.x },
                ["y"] = { AsFloat = pos.y },
                ["z"] = { AsFloat = pos.z }
            };
            return jc;
        }

        public static JSONClass Serialize(Vector4 vec)
        {
            var jc = new JSONClass
            {
                ["x"] = { AsFloat = vec.x },
                ["y"] = { AsFloat = vec.y },
                ["z"] = { AsFloat = vec.z },
                ["w"] = { AsFloat = vec.w },
            };
            return jc;
        }

        public static JSONClass Serialize(CtrlSnapshot snapshot)
        {
            if (snapshot == null || snapshot.name == string.Empty) return null;

            return new JSONClass
            {
                ["name"] = snapshot.name,
                ["currentPositionState"] = snapshot.currentPositionState.ToString(),
                ["currentRotationState"] = snapshot.currentRotationState.ToString(),
                ["RBHoldPositionSpring"] = snapshot.RBHoldPositionSpring.ToString(),
                ["RBHoldRotationSpring"] = snapshot.RBHoldRotationSpring.ToString(),
                ["localPosition"] = Serialize(snapshot.localPosition),
                ["localRotation"] = Serialize(snapshot.localRotation),
                ["worldMatrix"] = Serialize(snapshot.worldMatrix)
            };
        }

        public static Matrix4x4 DeserializeMatrix4x4(JSONClass jc)
        {
            return new Matrix4x4
            (
                DeserializeVector4(jc["0"] as JSONClass),
                DeserializeVector4(jc["1"] as JSONClass),
                DeserializeVector4(jc["2"] as JSONClass),
                DeserializeVector4(jc["3"] as JSONClass)
            );
        }

        public static Quaternion DeserializeQuaternion(JSONClass jc)
        {
            return new Quaternion
            (
                jc["x"].AsFloat,
                jc["y"].AsFloat,
                jc["z"].AsFloat,
                jc["w"].AsFloat
            );
        }

        public static Vector3 DeserializeVector3(JSONClass jc)
        {
            return new Vector3
            (
                jc["x"].AsFloat,
                jc["y"].AsFloat,
                jc["z"].AsFloat
            );
        }
        public static Vector4 DeserializeVector4(JSONClass jc)
        {
            return new Vector4
            (
                jc["x"].AsFloat,
                jc["y"].AsFloat,
                jc["z"].AsFloat,
                jc["w"].AsFloat
            );
        }

        public static CtrlSnapshot DeserializeCtrlSnapshot(JSONClass jc)
        {

            try
            {
                var settings = new CtrlSnapshot
                {
                    name = jc["name"].Value,
                    currentPositionState = jc["currentPositionState"].Value,
                    currentRotationState = jc["currentRotationState"].Value,
                    RBHoldPositionSpring = jc["RBHoldPositionSpring"].AsFloat,
                    RBHoldRotationSpring = jc["RBHoldRotationSpring"].AsFloat,
                    localPosition = DeserializeVector3(jc["localPosition"] as JSONClass),
                    localRotation = DeserializeQuaternion(jc["localRotation"] as JSONClass),
                    worldMatrix = DeserializeMatrix4x4(jc["worldMatrix"] as JSONClass)
                };

                return settings;
            }
            catch (Exception e)
            {
                SuperController.LogError($"{e}");
                return null;
            }
        }
    }

    public class TargetManager
    {
        private readonly bool __DEBUG__ = false;
        private JSONStorableStringChooser _targetAtomJSON;
        private JSONStorableStringChooser _targetJSON;
        private JSONStorableFloat _posOffsetXJSON;
        private JSONStorableFloat _posOffsetYJSON;

        private Atom _targetAtom;
        private static readonly List<string> _targetChoices = new List<string> { "Anus", "Mouth", "Vagina" };
        private static readonly Dictionary<string, List<string>> _targetNames = new Dictionary<string, List<string>>
        {
            {"Anus", new List<string> {"LabiaTrigger", "DeepVaginaTrigger"}},
            {"Mouth", new List<string> {"LipTrigger", "MouthTrigger"}},
            {"Vagina", new List<string> {"VaginaTrigger", "DeepVaginaTrigger"}}
        };
        private Rigidbody _target1;
        private Rigidbody _target2;
        public Vector3 TargetPos1 => _target1.transform.position - OffsetDirY * (_targetJSON.val == "Anus" ? 0.04f : _targetJSON.val == "Mouth" ? 0.01f : 0f) + _posOffset;
        public Vector3 TargetPos2 => _target2.transform.position + _posOffset;
        public Vector3 PosOffset => _posOffset;

        public Rigidbody TargetRB1 => _target1;
        public Rigidbody TargetRB2 => _target2;
        public Vector3 OffsetDirX => _target1.transform.right;
        public Vector3 OffsetDirY => _targetJSON.val == "Mouth" ? _target1.transform.up : _target1.transform.forward;
        private Vector3 _posOffset => (OffsetDirX * _posOffsetXJSON.val + OffsetDirY * _posOffsetYJSON.val) / 10;

        public JSONStorableStringChooser.SetStringCallback OnTargetAtomChange;
        public JSONStorableStringChooser.SetStringCallback OnTargetChange;

        public TargetManager(JSONStorableStringChooser.SetStringCallback OnTargetAtomChange = null, JSONStorableStringChooser.SetStringCallback OnTargetChange = null)
        {
            this.OnTargetAtomChange = OnTargetAtomChange ?? (_ => { });
            this.OnTargetChange = OnTargetChange ?? (_ => { });
        }

        private void _onTargetAtomChange(string atomUID)
        {
            if (__DEBUG__) SuperController.LogMessage($"Loading target Atom {atomUID}");

            if (atomUID == null || atomUID == string.Empty)
            {
                _targetAtom = null;
                _target1 = null;
                _target2 = null;
                return;
            }

            _targetAtom = SuperController.singleton.GetAtomByUid(atomUID);
            if (__DEBUG__) SuperController.LogMessage($"Loaded target atom {atomUID} ({_targetAtom})");

            OnTargetAtomChange(atomUID);

            _onTargetChange(_targetJSON.val);
        }

        private void _onTargetChange(string targetName)
        {
            if (__DEBUG__) SuperController.LogMessage($"Loading target {targetName}");

            if (_targetAtom != null && _targetNames.ContainsKey(targetName))
            {
                try
                {
                    _target1 = _targetAtom.rigidbodies.First(rb => rb.name == _targetNames[targetName][0]);
                    _target2 = _targetAtom.rigidbodies.First(rb => rb.name == _targetNames[targetName][1]);
                    OnTargetChange(targetName);
                    return;
                }
                catch (Exception e)
                {
                    SuperController.LogError($"Failed to load target rigidbodies: {e}");
                }
            }

            _target1 = null;
            _target2 = null;
        }

        public void SetupUI(MVRScript script, bool targetSelect = true, bool targetAdjust = true)
        {
            if (targetSelect)
            {
                _targetAtomJSON = new JSONStorableStringChooser("TargetAtom", Scene_Utils.PersonNames.FindAll(n => n != script.containingAtom.name), string.Empty, "<size=28><color=#000000>Target Atom</color></size>");
                script.RegisterStringChooser(_targetAtomJSON);
                _targetAtomJSON.setCallbackFunction += _onTargetAtomChange;
                script.CreateScrollablePopup(_targetAtomJSON).popup.onOpenPopupHandlers += () => { _targetAtomJSON.choices = Scene_Utils.PersonNames.FindAll(n => n != script.containingAtom.name); };

                _targetJSON = new JSONStorableStringChooser("Target", _targetChoices, _targetChoices.Last(), "<size=28><color=#000000>Target</color></size>");
                script.RegisterStringChooser(_targetJSON);
                _targetJSON.setCallbackFunction += _onTargetChange;
                script.CreateScrollablePopup(_targetJSON).popup.onOpenPopupHandlers += () => { _targetJSON.choices = _targetChoices; };
            }

            if (targetAdjust)
            {
                _posOffsetXJSON = UI_Utils.SetupSlider(script, "<size=24>Target Adj. Left-Right</size>", 0f, -1f, 1f);
                _posOffsetYJSON = UI_Utils.SetupSlider(script, "<size=24>Target Adj. Down-Up</size>", 0f, -1f, 1f);
            }
        }

    }

    public static class UI_Utils
    {
        public static JSONStorableFloat SetupSlider(MVRScript script, string name, float defaultValue, float minValue, float maxValue, bool constrain = true, bool rightSide = false)
        {
            JSONStorableFloat storable = new JSONStorableFloat(name, defaultValue, minValue, maxValue, constrain)
            {
                storeType = JSONStorableParam.StoreType.Full
            };
            script.CreateSlider(storable, rightSide);
            script.RegisterFloat(storable);
            return storable;
        }

        public static JSONStorableBool SetupToggle(MVRScript script, string label, bool defaultValue, bool rightSide = false)
        {
            JSONStorableBool storable = new JSONStorableBool(label, defaultValue)
            {
                storeType = JSONStorableParam.StoreType.Full
            };
            script.CreateToggle(storable, rightSide);
            script.RegisterBool(storable);
            return storable;
        }

        // Stolen from MacGruber
        public static UIDynamicButton SetupButton(MVRScript script, string label, UnityEngine.Events.UnityAction callback, bool rightSide = false)
        {
            UIDynamicButton button = script.CreateButton(label, rightSide);
            button.button.onClick.AddListener(callback);
            return button;
        }
    }

}